@@ -9,6 +9,7 @@ warnings.filterwarnings("ignore")
 import os
 import sys
 import json
+import pickle
 import shutil
 import inspect
 import configparser
@@ -25,9 +26,11 @@ from termcolor import colored
 ### Logging ###
 import logging
 import logzero
-from logzero import logger, LogFormatter
+from rich.logging import RichHandler
 
 ### Data Handling ###
+import numpy as np
+
 # import pandas as pd
 
 ### Hyperparameter Optimization ###
@@ -35,11 +38,16 @@ from mango import scheduler, Tuner
 
 ### MoonLine ###
 from linechart import Tearsheet, DailyPerformance, AggregateDailyPerformance
-from moonline import main as moonline_main, get_config as moonline_get_config
+from moonline import (
+    main as moonline_main,
+    get_config as moonline_get_config,
+    get_strategies_dir as moonline_get_strategies_dir,
+)
 
 
 space_x_config = {}
 moonline_config = {}
+best_param_hash = None
 moonline_config_custom = {}
 
 
@@ -83,7 +91,7 @@ def get_config(config_file, args=[]):
             else:
                 section, value = None, None
         elif section and value:
-            config_object_custom[section][value] = item
+            config_object_custom[section.lower()][value] = item
 
     errors = []
 
@@ -114,6 +122,12 @@ def get_config(config_file, args=[]):
     else:
         config_object["Output"]["best_tearsheet"] = None
 
+    results = config_object["Output"].get("results", None)
+    if results:
+        config_object["Output"]["results"] = os.path.abspath(config_object["Output"]["results"])
+    else:
+        config_object["Output"]["results"] = None
+
     config_object["Parameters"]["path"] = os.path.abspath(config_object["Parameters"]["path"])
     if not os.path.isdir(config_object["Parameters"]["path"]):
         errors.append(("[Parameters]->path", "Not a directory", config_object["Parameters"]["path"]))
@@ -238,7 +252,11 @@ def moonline_objective(**kwargs):
 
     logzero.loglevel(logging.CRITICAL)
 
-    moonline_main(config, config_custom)
+    try:
+        moonline_main(config, config_custom)
+    except Exception as e:
+        logger.exception(f"Ran into an exception during the backtest, returning 0: {e}")
+        return 0
 
     logzero.loglevel(logging.INFO)
 
@@ -253,14 +271,32 @@ def moonline_objective(**kwargs):
 
     score = (perf_agg.cagr - (perf_agg.cagr * perf_agg.max_drawdown)) + perf_agg.sharpe
 
-    # logger.info(f"CAGR: {perf_agg.cagr * 100:.2f}%")
-    # logger.info(f"Sharpe: {perf_agg.sharpe}")
-    # logger.info(f"Drawdown: {perf_agg.max_drawdown * 100:.2f}%")
-    # logger.info(f"Score: {score}")
+    if np.isnan(score) or str(score) == "nan":
+        score = 0
+
+    logger.debug(
+        f"Score: {score:.2f} | Parameters: {kwargs} | CAGR: {perf_agg.cagr * 100:.2f}% | Sharpe: {perf_agg.sharpe} | Drawdown: {perf_agg.max_drawdown * 100:.2f}%"
+    )
 
     return score
 
 
+def early_stopping(results):
+    global params, best_param_hash
+
+    params_hash = make_hash_md5(results["best_params"])
+
+    if params_hash != best_param_hash:
+        logger.info(f"Best Parameters: {results['best_params']}")
+        logger.info(f"Best Objective: {results['best_objective']}")
+        best_param_hash = params_hash
+
+    if hasattr(params, "early_stopping"):
+        return params.early_stopping(results)
+
+    return False
+
+
 @click.command(context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
 @click.option(
     "-c",
@@ -273,15 +309,16 @@ def moonline_objective(**kwargs):
 @click.option("-v", "--verbose", is_flag=True)
 @click.argument("args", nargs=-1, type=click.UNPROCESSED)
 def cli(config_file, verbose, args):
-    global space_x_config, moonline_config, moonline_config_custom
+    global space_x_config, moonline_config, moonline_config_custom, params, logger
 
-    formatter = LogFormatter(
-        fmt="%(color)s[%(levelname)1.1s %(asctime)s]%(end_color)s %(message)s",
-        datefmt="%H:%M:%S",
+    logging.basicConfig(
+        level=logging.DEBUG if verbose else logging.INFO,
+        format="%(message)s",
+        datefmt="[%X]",
+        handlers=[RichHandler(rich_tracebacks=True)],
     )
-    logzero.formatter(formatter)
 
-    logzero.loglevel(logging.DEBUG if verbose else logging.INFO)
+    logger = logging.getLogger("rich")
 
     space_x_config, _ = get_config(config_file)
 
@@ -305,13 +342,30 @@ def cli(config_file, verbose, args):
     # Fill empty system section
     if "System" not in moonline_config:
         moonline_config["System"] = {}
-    # Disable shutil.rmtree during strategy copy (this conflicts when using parallel processing)
-    moonline_config["System"]["clear_tmp_dir"] = False
+    # Disable strategy copying and folder clearing (causes conflicts with parallel processing)
+    moonline_config["System"]["read_only"] = True
     # Enable quiet mode to reduce log spam
     moonline_config["System"]["quiet"] = True
 
+    # Copy strategies
+    strategy_files = glob(os.path.join(moonline_config["Strategy"]["path"], "**", "*.py"), recursive=True)
+    tmp_dir = moonline_get_strategies_dir()
+    if os.path.isdir(tmp_dir):
+        shutil.rmtree(tmp_dir)
+    os.makedirs(tmp_dir, exist_ok=True)
+    for file in strategy_files:
+        shutil.copyfile(file, os.path.join(tmp_dir, os.path.basename(file)))
+
     logger.info(f"Running Hyperparameter Optimization with {space_x_config['SpaceX']['iterations']} iterations")
-    tuner = Tuner(param_space, moonline_objective, {"num_iteration": space_x_config["SpaceX"]["iterations"]})
+    tuner = Tuner(
+        param_space,
+        moonline_objective,
+        {
+            "initial_random": 4,
+            "num_iteration": space_x_config["SpaceX"]["iterations"],
+            "early_stopping": early_stopping,
+        },
+    )
 
     results = tuner.maximize()
 
@@ -331,6 +385,10 @@ def cli(config_file, verbose, args):
     for params_tried, score in sorted(zip(results["params_tried"], results["objective_values"]), key=lambda x: x[1]):
         logger.info(f"{params_tried}: {score}")
 
+    if space_x_config["Output"]["results"]:
+        with open(space_x_config["Output"]["results"], "wb") as f:
+            pickle.dump(results, f)
+
 
 if __name__ == "__main__":
     cli()
